// Copyright 2000-2005 the Contributors, as shown in the revision logs.
// Licensed under the Apache Public Source License 2.0 ("the License").
// You may not use this file except in compliance with the License.

package org.ibex.io;

import java.io.*;
import java.net.*;
import java.util.*;
import java.util.zip.*;
import org.ibex.util.*;

/** plays the role of InputStream, OutputStream, Reader and Writer, with logging and unchecked exceptions */
public class Stream {

    protected final In in;
    protected final Out out;
    private         String newLine = "\r\n";
    private         Stream in_next = null;

    public Stream append(String in_next) { return appendStream(new Stream(in_next)); }
    public Stream appendStream(Stream in_next) {
        if (this.in_next != null)
            this.in_next.appendStream(in_next);
        else
            this.in_next = in_next;
        return this;
    }

    public void transcribe(Stream out) { transcribe(out, false); }
    public void transcribe(Stream out, boolean close) {
        try {
            byte[] buf = new byte[1024];
            while(true) {
                int numread = in.read(buf, 0, buf.length);
                if (numread==-1) { in.close(); break; }
                out.out.write(buf, 0, numread);
            }
	    if (close) out.close();
        } catch (IOException ioe) { ioe(ioe); }
    }

    public void transcribe(StringBuffer out) {
        //try {
            char[] buf = new char[1024];
            while(true) {
                int numread = in.read(buf, 0, buf.length);
                if (numread==-1) { in.close(); return; }
                out.append(buf, 0, numread);
            }
            //} catch (IOException ioe) { ioe(ioe); }
    }

    public static int countLines(Stream s) {
        int ret = 0;
        while(s.readln() != null) ret++;
        s.close();
        return ret;
    }

    public  Stream(byte[] b, int off, int len) { this.in = new Stream.In(new ByteArrayInputStream(b, off, len)); this.out=null; }
    public  Stream(InputStream in)                   { this.in = new Stream.In(in); this.out = null; }
    public  Stream(                OutputStream out) { this.in = null;              this.out = new Stream.Out(out); }
    public  Stream(InputStream in, OutputStream out) { this.in = new Stream.In(in); this.out = new Stream.Out(out); }
    public  Stream(String s)                         { this(new ByteArrayInputStream(s.getBytes())); }
    public  Stream(File f)                           {
        try { this.in = new Stream.In(new FileInputStream(f)); } catch (IOException e) { ioe(e); throw new Error(); }
        this.out = null;
    }
    public  Stream(Socket s) {
        try { this.in = new Stream.In(s.getInputStream());    } catch (IOException e) { ioe(e); throw new Error(); }
        try { this.out = new Stream.Out(s.getOutputStream()); } catch (IOException e) { ioe(e); throw new Error(); }
    }

    static int ioe(IOException e) {
        if (e instanceof SocketException && e.toString().indexOf("Connection reset")!=-1)
            throw new Closed(e.getMessage());
        throw new StreamException(e);
    }
    public static class StreamException extends RuntimeException {
        public StreamException(Exception e) { super(e); }
        public StreamException(String s)    { super(s); }
    }
    public static class EOF             extends StreamException  { public EOF() { super("End of stream"); } }
    public static class Closed          extends StreamException  { public Closed(String s) { super(s); } }


    public char   peekc()                          {
        flush();
        return in.getc(true);
    }
    public char   getc()                           {
        flush();
        char ret = in.getc(false);
        return ret;
    }   
    public String readln()                         {
        flush();
        String s = in.readln();
        return s;
    }
    
    public void   print(String s) {
        out.write(s);
        flush();
    }
    public void   println(String s) {
        out.write(s);
        out.write(newLine);
        flush();
    }
    public void   flush()                          {
        if (out != null) try { out.w.flush(); } catch(IOException e) { ioe(e); }
    }
    public int    read(byte[] b, int off, int len) {
        flush();
        int ret = in.readBytes(b, off, len);
        nnl = false;
        return ret;
    }
    public int    read(char[] c, int off, int len) {
        flush();
        int ret = in.read(c, off, len);
        return ret;
    }

    public void   unread(String s)                 { in.unread(s); }

    /** should not throw exceptions */
    public void   close()                          { try { if (in!=null) in.close(); } finally { if (out!=null) out.close(); } }
    public void   setNewline(String s)             { newLine = s; }


    /** dumps the connection log into a file */
    private boolean nnl = false;

    private static class Out extends BufferedOutputStream {
        private Writer w = new BufferedWriter(new OutputStreamWriter(this));
        public Out(OutputStream out) { super(out); }
        public void close() { try { super.close(); } catch (Exception e) { Log.error(this, e); } }
        public void write(String s) { try { w.write(s); } catch (IOException e) { ioe(e); } }
    }

    private class In extends InputStream {
        public final Reader reader = new InputStreamReader(this);
        private /*final*/ InputStream orig;
        public In(InputStream in) { orig = in; }

        char[] cbuf = new char[8192];
        int cstart = 0;
        int cend = 0;

        byte[] buf = new byte[8192];
        int start = 0;
        int end = 0;

        boolean flushing = false;

        public int available() { return flushing ? 0 : (end - start); }
        public void close() { try {
            if (orig!=null) orig.close(); 
            if (in_next != null) in_next.close();  // FIXME: correct?
        } catch (Exception e) { Log.error(this, e); } }

        public char getc(boolean peek) { try {
            if (cstart == cend) {
                cstart = 0;
                cend = reader.read(cbuf, 0, cbuf.length);
                if (cend == -1) {
                    reader.close();
                    cend = cstart;
                    if (in_next == null) throw new EOF();
                    return getc(peek);
                }
            }
            return peek ? cbuf[cstart] : cbuf[cstart++];
        } catch (IOException e) { return (char)ioe(e); } }

        public String readln() { try {
            while(true) {
                for(int i=cstart; i<cend; i++)
                    if (cbuf[i] == '\n') {
                        // this should (in theory) handle CR, LF,
                        // CRLF, and LFCR properly, assuming that the
                        // file consistently uses the same ending
                        // throughout.
                        int begin = cstart;
                        int len = i-cstart;
                        cstart = i+1;
                        if (cbuf[begin] == '\r') { begin++; len--; }
                        while (len > 0 && cbuf[begin+len-1] == '\r') { len--; }
                        return new String(cbuf, begin, len);
                    }
                ensurec(256);
                int numread = reader.read(cbuf, cend, cbuf.length - cend);
                if (numread == -1) {
                    reader.close();
                    if (cstart == cend) return null;
                    String ret = new String(cbuf, cstart, cend-cstart);
                    cstart = cend = 0;
                    return ret;
                }
                cend += numread;
            }
        } catch (IOException e) { ioe(e); return null; } }

        public int read(char[] c, int pos, int len) { try {
            if (cstart == cend) {
                cstart = 0;
                cend = reader.read(cbuf, 0, cbuf.length);
                if (cend == -1) { reader.close(); cend = cstart; return -1; }
            }
            if (len > cend - cstart) len = cend - cstart;
            System.arraycopy(cbuf, cstart, c, pos, len);
            cstart += len;
            return len;
        } catch (IOException e) { ioe(e); return -1; } }

        public int readBytes(byte[] b, int pos, int len) { flushchars(); return read(b, pos, len); }
        public int read() { byte[] b = new byte[1]; if (read(b, 0, 1) == -1) return -1; return (int)b[0]; } 
        public int read(byte[] b, int pos, int len) { try {
            if (start == end) {
                start = 0;
                end = orig.read(buf, 0, buf.length);
                if (end == -1) {
                    orig.close();
                    end = start;
                    if (in_next==null) return -1;
                    return in_next.read(b, pos, len);
                }
            }
            if (len > end - start) len = end - start;
            System.arraycopy(buf, start, b, pos, len);
            start += len;
            return len;
        } catch (IOException e) { ioe(e); return -1; } }

        private void growc(int s){char[] cbuf2=new char[cbuf.length+s*2];System.arraycopy(cbuf,0,cbuf2,0,cbuf.length);cbuf=cbuf2; }
        private void shiftc() {
            char[] cbuf2 = new char[cbuf.length];
            System.arraycopy(cbuf, cstart, cbuf2, 0, cend-cstart);
            cend -= cstart;
            cstart = 0;
            cbuf = cbuf2;
        }
        private void ensurec(int space) { if (cend-cstart+space>cbuf.length) growc(space); if (cend+space>cbuf.length) shiftc(); }

        private void growb(int s) { byte[] buf2 = new byte[buf.length+s*2]; System.arraycopy(buf,0,buf2,0,buf.length); buf=buf2; }
        private void shiftb() { System.arraycopy(buf, start, buf, 0, end-start); end -= start; start = 0; }
        private void ensureb(int space) { if (end-start+space>buf.length) growb(space); if (end+space>buf.length) shiftb(); }
        private void ensureb2(int space) { if (end-start+space>buf.length) growb(space); if (start<space) unshiftb(); }
        private void unshiftb() {
            System.arraycopy(buf,start,buf,buf.length-(end-start),end-start);start=buf.length-(end-start);end=buf.length; }

        public  void unread(String s) { ensurec(s.length()); s.getChars(0, s.length(), cbuf, cend); cend += s.length(); }

        private void flushchars() {
            try {
                flushing = true;
                for(; reader.ready(); reader.read(cbuf, cend++, 1)) ensurec(10);
                if (cend>cstart)
                    unreader.write(cbuf, cstart, cend-cstart);
                cstart = cend = 0;
                unreader.flush();
            } catch (IOException e) { ioe(e);
            } finally { flushing = false; }
        }

        Writer unreader = new OutputStreamWriter(new InOutputStream());
	private class InOutputStream extends OutputStream {
	    public void close() { }
	    public void write(int i) throws IOException { byte[] b = new byte[1]; b[0] = (byte)i; write(b, 0, 1); }
	    public void write(byte[] b) throws IOException { write(b, 0, b.length); }
	    public void write(byte[] b, int p, int l) {
		ensureb2(l);
		System.arraycopy(b, p, buf, start-l, l);
		start -= l;
	    }
	}
    }

}
